import Eto.Drawing  # type: ignore
import Eto.Forms  # type: ignore
import Rhino  # type: ignore
import Rhino.UI  # type: ignore
import System  # type: ignore


class SplashForm(Eto.Forms.Dialog[bool]):
    def __init__(self, title, url, width=800, height=450):
        super().__init__()

        self.Title = title
        self.Padding = Eto.Drawing.Padding(0)
        self.Resizable = False
        self.ClientSize = Eto.Drawing.Size(width, height)
        self.WindowStyle = Eto.Forms.WindowStyle.NONE

        webview = Eto.Forms.WebView()
        webview.Size = Eto.Drawing.Size(width, height)
        webview.Url = System.Uri(url)
        webview.BrowserContextMenuEnabled = False
        webview.DocumentLoading += self.action

        layout = Eto.Forms.DynamicLayout()
        layout.BeginVertical()
        layout.AddRow(webview)
        layout.EndVertical()
        self.Content = layout

    def action(self, sender, e):
        if e.Uri.Scheme == "action" and e.Uri.Host == "close":
            self.Close()

    def show(self):
        return self.ShowModal(Rhino.UI.RhinoEtoApp.MainWindow)



import rhinoscriptsyntax as rs  # type: ignore


def confirm(message):
    result = rs.MessageBox(message, buttons=4 | 32 | 256 | 0, title="Confirmation")
    return result == 6


def warn(message):
    return rs.MessageBox(message, title="Warning")



import Eto.Drawing  # type: ignore
import Eto.Forms  # type: ignore
import Rhino  # type: ignore
import Rhino.UI  # type: ignore


class InfoForm(Eto.Forms.Dialog[bool]):
    def __init__(self, text, title="Info", width=800, height=500):
        super().__init__()

        self.Title = title
        self.Padding = Eto.Drawing.Padding(0)
        self.Resizable = True
        self.MinimumSize = Eto.Drawing.Size(0.5 * width, 0.5 * height)
        self.ClientSize = Eto.Drawing.Size(width, height)
        textarea = Eto.Forms.TextArea()
        textarea.Text = text
        textarea.ReadOnly = True
        layout = Eto.Forms.DynamicLayout()
        layout.BeginVertical(Eto.Drawing.Padding(12, 12, 12, 0), Eto.Drawing.Size(0, 0), True, True)
        layout.AddRow(textarea)
        layout.EndVertical()
        layout.BeginVertical(Eto.Drawing.Padding(12, 12, 12, 18), Eto.Drawing.Size(6, 0), False, False)
        layout.AddRow(None, self.ok)
        layout.EndVertical()
        self.Content = layout

    @property
    def ok(self):
        self.DefaultButton = Eto.Forms.Button()
        self.DefaultButton.Text = "OK"
        self.DefaultButton.Click += self.on_ok
        return self.DefaultButton

    def on_ok(self, sender, event):
        self.Close(True)

    def show(self):
        return self.ShowModal(Rhino.UI.RhinoEtoApp.MainWindow)



import Eto.Drawing  # type: ignore
import Eto.Forms  # type: ignore
import Rhino  # type: ignore
import Rhino.UI  # type: ignore

from compas import json_dump
from compas.data import Data
from compas.datastructures import Mesh

from .filesystem import FileForm
from .meshdata import MeshDataForm
from .settings import SettingsForm


class ObjCell(Eto.Forms.CustomCell):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent

    def OnGetIdentifier(self, args):
        return str(args.Row)


class ObjNameCell(ObjCell):
    def OnCreateCell(self, args):
        def on_text_changed(sender, e):
            obj.name = sender.Text

        obj = args.Item.GetValue(0)
        control = Eto.Forms.TextBox()
        control.Text = obj.name
        control.ShowBorder = False
        control.TextChanged += on_text_changed
        return control


class ObjTypeCell(ObjCell):
    def OnCreateCell(self, args):
        obj = args.Item.GetValue(0)
        control = Eto.Forms.Label(
            Text=obj.item.__class__.__name__,
            TextAlignment=Eto.Forms.TextAlignment.Left,
        )
        return control


class VisibleCell(ObjCell):
    def OnCreateCell(self, args):
        def on_click(sender, e):
            obj.visible = sender.Checked
            self.parent.scene.update()

        obj = args.Item.GetValue(0)
        control = Eto.Forms.CheckBox()
        control.Checked = obj.visible
        control.CheckedChanged += on_click
        return control


class SettingsCell(ObjCell):
    def OnCreateCell(self, args):
        def on_click(sender, e):
            form = SettingsForm(obj.settings)
            if form.ShowModal(self.parent):
                obj.settings.update(form.settings)
                self.parent.scene.update()

        obj = args.Item.GetValue(0)
        if isinstance(obj.item, Mesh):
            control = Eto.Forms.Button()
            control.Text = "Settings"
            control.BackgroundColor = Eto.Drawing.Colors.Gray
            control.Click += on_click
            return control


class ItemCell(ObjCell):
    def OnCreateCell(self, args):
        def on_click(sender, e):
            form = MeshDataForm(obj.item)
            if form.ShowModal(self.parent):
                self.parent.scene.update()

        obj = args.Item.GetValue(0)
        if isinstance(obj.item, Mesh):
            control = Eto.Forms.Button()
            control.Text = "Data"
            control.BackgroundColor = Eto.Drawing.Colors.Gray
            control.Click += on_click
            return control


# class DefaultAttributesCell(ObjCell):
#     def OnCreateCell(self, args):
#         def on_click(sender, e):
#             form = DefaultAttributesForm(obj.item)
#             if form.ShowModal(self.parent):
#                 self.parent.scene.update()

#         obj = args.Item.GetValue(0)
#         if isinstance(obj.item, Mesh):
#             control = Eto.Forms.Button()
#             control.Text = "Default Attributes"
#             control.BackgroundColor = Eto.Drawing.Colors.Gray
#             control.Click += on_click
#             return control


class ExportCell(ObjCell):
    def OnCreateCell(self, args):
        def on_click(sender, e):
            path = FileForm.save(basename=obj.name + ".json")
            if not path:
                return
            json_dump(obj.item, path)

        obj = args.Item.GetValue(0)
        if isinstance(obj.item, Data):
            control = Eto.Forms.Button()
            control.Text = "Export"
            control.BackgroundColor = Eto.Drawing.Colors.Gray
            control.Click += on_click
            return control


class RemoveCell(ObjCell):
    def OnCreateCell(self, args):
        def on_click(sender, e):
            result = Eto.Forms.MessageBox.Show(
                "Are you sure you want to remove this object?",
                "Confirm",
                Eto.Forms.MessageBoxButtons.YesNo,
            )
            if result == Eto.Forms.DialogResult.Yes:
                self.parent.scene.remove(obj)
                self.parent.scene.update()
                self.parent.map_objects()

        obj = args.Item.GetValue(0)
        control = Eto.Forms.Button()
        control.Text = "Remove"
        control.BackgroundColor = Eto.Drawing.Colors.Gray
        control.Click += on_click
        return control


class ActiveCell(ObjCell):
    def OnCreateCell(self, args):
        def on_click(sender, e):
            self.parent.scene.active_object = obj
            for _obj, _control in self.parent.active_controls:
                if _obj != obj:
                    _control.Checked = self.parent.scene.active_object == _obj
            # print("active object:", self.parent.scene.active_object)

        obj = args.Item.GetValue(0)
        control = Eto.Forms.RadioButton()
        control.Checked = obj.active
        control.MouseDown += on_click
        self.parent.active_controls.append((obj, control))

        return control


class SceneObjectsForm(Eto.Forms.Dialog[bool]):
    def __init__(self, scene, title="Scene Objects", width=800, height=500):
        super().__init__()

        self.Title = title
        self.Padding = Eto.Drawing.Padding(0)
        self.Resizable = True
        self.MinimumSize = Eto.Drawing.Size(0.5 * width, 0.5 * height)
        self.ClientSize = Eto.Drawing.Size(width, height)

        self.scene = scene
        self.table = Eto.Forms.TreeGridView()
        self.table.ShowHeader = True

        column = Eto.Forms.GridColumn()
        column.HeaderText = "Object"
        column.Editable = False
        column.DataCell = ObjTypeCell(self)
        self.table.Columns.Add(column)

        column = Eto.Forms.GridColumn()
        column.HeaderText = "Name"
        column.Editable = False
        column.DataCell = ObjNameCell(self)
        self.table.Columns.Add(column)

        # column = Eto.Forms.GridColumn()
        # column.HeaderText = "Active"
        # column.Editable = False
        # column.DataCell = ActiveCell(self)
        # self.table.Columns.Add(column)

        column = Eto.Forms.GridColumn()
        column.HeaderText = "Visible"
        column.Editable = False
        column.DataCell = VisibleCell(self)
        self.table.Columns.Add(column)

        column = Eto.Forms.GridColumn()
        column.HeaderText = ""
        column.Editable = False
        column.DataCell = SettingsCell(self)
        self.table.Columns.Add(column)

        # column = Eto.Forms.GridColumn()
        # column.HeaderText = ""
        # column.Editable = False
        # column.DataCell = DefaultAttributesCell(self)
        # self.table.Columns.Add(column)

        column = Eto.Forms.GridColumn()
        column.HeaderText = ""
        column.Editable = False
        column.DataCell = ItemCell(self)
        self.table.Columns.Add(column)

        column = Eto.Forms.GridColumn()
        column.HeaderText = ""
        column.Editable = False
        column.DataCell = ExportCell(self)
        self.table.Columns.Add(column)

        # column = Eto.Forms.GridColumn()
        # column.HeaderText = ""
        # column.Editable = False
        # column.DataCell = RemoveCell(self)
        # self.table.Columns.Add(column)

        self.map_objects()

        layout = Eto.Forms.DynamicLayout()
        layout.BeginVertical(Eto.Drawing.Padding(0, 0, 0, 0), Eto.Drawing.Size(0, 0), True, True)
        layout.AddRow(self.table)
        layout.EndVertical()

        self.Content = layout
        self.active_controls = []

    def map_objects(self):
        def add_items(parent, objects):
            for obj in objects:
                item = Eto.Forms.TreeGridItem()
                item.Values = (obj,)
                if obj.children:
                    add_items(item.Children, obj.children)
                parent.Add(item)

        collection = Eto.Forms.TreeGridItemCollection()
        root_objects = [obj for obj in self.scene.objects if obj.parent is None]
        add_items(collection, root_objects)
        self.table.DataStore = collection

    def show(self):
        return self.ShowModal(Rhino.UI.RhinoEtoApp.MainWindow)



from ast import literal_eval

import Eto.Drawing  # type: ignore
import Eto.Forms  # type: ignore
import Rhino  # type: ignore
import Rhino.UI  # type: ignore

from compas.datastructures import Mesh


class MeshDataForm(Eto.Forms.Dialog[bool]):
    """Form for working with mesh data.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        A COMPAS mesh.
    title : str, optional
        Title of the dialog.
    width : int, optional
        Window width.
    height : int, optional
        Window height.
    exclude_vertex_attr : list[str], optional
        The names of vertex attributes to exclude.
    exclude_edge_attr : list[str], optional
        The names of edge attributes to exclude.
    exclude_face_attr : list[str], optional
        The names of face attributes to exclude.

    """

    def __init__(
        self,
        mesh: Mesh,
        title="Mesh Data",
        width=800,
        height=800,
        excluded_vertex_attr=None,
        excluded_edge_attr=None,
        excluded_face_attr=None,
    ):
        super().__init__()

        self.mesh = mesh
        self.excluded_vertex_attr = excluded_vertex_attr
        self.excluded_edge_attr = excluded_edge_attr
        self.excluded_face_attr = excluded_face_attr

        self.Title = title
        self.Padding = Eto.Drawing.Padding(0)
        self.Resizable = True
        self.MinimumSize = Eto.Drawing.Size(0.5 * width, 0.5 * height)
        self.ClientSize = Eto.Drawing.Size(width, height)

        self.vertexpage = Page(
            title="Vertices",
            defaults=self.mesh.default_vertex_attributes,
            keys=list(self.mesh.vertices()),
            valuefunc=self.mesh.vertex_attribute,
            excluded=self.excluded_vertex_attr,
        )
        self.edgepage = Page(
            title="Edges",
            defaults=self.mesh.default_edge_attributes,
            keys=list(self.mesh.edges()),
            valuefunc=self.mesh.edge_attribute,
            excluded=self.excluded_edge_attr,
        )
        self.facepage = Page(
            title="Faces",
            defaults=self.mesh.default_face_attributes,
            keys=list(self.mesh.faces()),
            valuefunc=self.mesh.face_attribute,
            excluded=self.excluded_face_attr,
        )
        control = Eto.Forms.TabControl()
        control.TabPosition = Eto.Forms.DockPosition.Top
        control.Pages.Add(self.vertexpage.widget)
        control.Pages.Add(self.edgepage.widget)
        control.Pages.Add(self.facepage.widget)

        layout = Eto.Forms.DynamicLayout()
        layout.BeginVertical(Eto.Drawing.Padding(0, 12, 0, 12), Eto.Drawing.Size(0, 0), True, True)
        layout.AddRow(control)
        layout.EndVertical()
        layout.BeginVertical(Eto.Drawing.Padding(12, 18, 12, 24), Eto.Drawing.Size(6, 0), False, False)
        layout.AddRow(None, self.ok, self.cancel)
        layout.EndVertical()
        self.Content = layout

    @property
    def ok(self):
        self.DefaultButton = Eto.Forms.Button(Text="OK")
        self.DefaultButton.Click += self.on_ok
        return self.DefaultButton

    @property
    def cancel(self):
        self.AbortButton = Eto.Forms.Button(Text="Cancel")
        self.AbortButton.Click += self.on_cancel
        return self.AbortButton

    def on_ok(self, sender, event):
        """Callback for the OK event."""
        try:
            self.vertexpage.process()
            self.edgepage.process()
            self.facepage.process()
        except Exception as e:
            print(e)
            self.Close(False)
        self.Close(True)

    def on_cancel(self, sender, event):
        """Callback for the CANCEL event."""
        self.Close(False)

    def show(self):
        """Show the form dialog."""
        return self.ShowModal(Rhino.UI.RhinoEtoApp.MainWindow)


class Page:
    """Wrapper for Eto tab pages.

    Parameters
    ----------
    title : str
        The title of the tab page.
    defaults : dict[str, Any]
        Default values of the data attributes.
    keys : list[int | tuple[int, int]]
        The identifiers of the mesh components.
    valuefunc : callable
        Function for retrieving the value of a named data attribute of a specific component.

    Attributes
    ----------
    names : list[str]
        The names of the attributes, in alphabetical order.
    public : list[str]
        The names of the editable attributes.
    private : list[str]
        The names of the read-only attributes.
    cols : list[dict]
        A list of dicts with each dict representing the properties of a data column.
    rows : list[list[str]]
        The data per mesh components corresponding to the columns.

    """

    def __init__(self, title, defaults, keys, valuefunc, excluded=None):
        self.defaults = defaults
        self.keys = keys
        self.valuefunc = valuefunc
        self.excluded = excluded or []
        self.table = Table(self.cols, self.rows)
        self.widget = Eto.Forms.TabPage()
        self.widget.Text = title
        if self.names:
            # replace this by a dynamic layout
            # with the first row an overview of the default attribute values
            # and second row the data table
            self.widget.Content = self.table.widget

    @property
    def names(self):
        return sorted([name for name in self.defaults.keys() if name not in self.excluded])

    @property
    def public(self):
        return sorted([name for name in self.names if not name.startswith("_")])

    @property
    def private(self):
        return sorted([name for name in self.names if name.startswith("_")])

    @property
    def cols(self):
        cols = [
            {
                "name": "ID",
                "is_editable": False,
                "is_checkbox": False,
                "precision": None,
            }
        ]

        for name in self.public:
            default = self.defaults[name]
            col = {
                "name": name,
                "is_editable": True,
                "is_checkbox": False,
                "precision": None,
            }
            if isinstance(default, bool):
                col["is_checkbox"] = True
            elif isinstance(default, float):
                col["precision"] = "3f"
            cols.append(col)

        for name in self.private:
            default = self.defaults[name]
            col = {
                "name": name,
                "is_editable": False,
                "is_checkbox": False,
                "precision": None,
            }
            if isinstance(default, bool):
                col["is_checkbox"] = True
            elif isinstance(default, float):
                col["precision"] = "3f"
            cols.append(col)

        return cols

    @property
    def rows(self):
        rows = []
        for key in self.keys:
            row = [str(key)]
            for col in self.cols[1:]:
                name = col["name"]
                checkbox = col["is_checkbox"]
                precision = col["precision"]
                value = self.valuefunc(key, name)
                if precision:
                    value = "{0:.{1}}".format(float(value), precision)
                elif not checkbox:
                    value = str(value)
                row.append(str(value))
            rows.append(row)
        return rows

    def process(self):
        """Process the data of the page."""
        for row in self.table.data:
            key = literal_eval(row[0])
            for index, col in enumerate(self.table.cols):
                name = col["name"]
                editable = col["is_editable"]
                precision = col["precision"]
                if not editable:
                    continue
                value = row[index]
                if isinstance(value, str):
                    try:
                        value = literal_eval(value)
                    except Exception as e:
                        print(key, name, value, e)
                    else:
                        if precision:
                            value = float(value)
                        self.valuefunc(key, name, value)
                else:
                    if precision:
                        value = float(value)
                    self.valuefunc(key, name, value)


class Table(object):
    """Wrapper for Eto grid view.

    Parameters
    ----------
    cols
    rows

    """

    def __init__(self, cols, rows):
        def on_cell_formatting(sender, e):
            try:
                if not e.Column.Editable:
                    e.ForegroundColor = Eto.Drawing.Colors.DarkGray
            except Exception as e:
                print(e)

        self.widget = Eto.Forms.GridView()
        self.widget.ShowHeader = True
        self.widget.DataStore = rows

        for i, col in enumerate(cols):
            column = Eto.Forms.GridColumn()
            column.HeaderText = col["name"]
            column.Editable = col["is_editable"]
            if col["is_checkbox"]:
                cell = Eto.Forms.CheckBoxCell(i)
                column.DataCell = cell
            else:
                cell = Eto.Forms.TextBoxCell(i)
                cell.AutoSelectMode = Eto.Forms.AutoSelectMode.OnFocus
                cell.VerticalAlignment = Eto.Forms.VerticalAlignment.Center
                cell.TextAlignment = Eto.Forms.TextAlignment.Right
                column.DataCell = cell

            self.widget.Columns.Add(column)

        self.widget.CellFormatting += on_cell_formatting

        self.cols = cols
        self.rows = rows
        self.data = self.widget.DataStore



from .value import Value


class NumberValue(Value):
    def _check_bounds(self, value):
        if self.minval is not None:
            if value < self.minval:
                raise ValueError("Value must be greater than {}".format(self.minval))

        if self.maxval is not None:
            if value > self.maxval:
                raise ValueError("Value must be less than {}".format(self.maxval))

    def check(self, value):
        super().check(value)
        self._check_bounds(value)

    @property
    def minval(self):
        return self._minval

    @minval.setter
    def minval(self, value):
        self._check_type(value)
        self._minval = value

    @property
    def maxval(self):
        return self._maxval

    @maxval.setter
    def maxval(self, value):
        self._check_type(value)
        self._maxval = value


class FloatValue(NumberValue):
    def __init__(self, value=0.0, minval=None, maxval=None, options=None):
        super().__init__(value, float, options=options)
        if minval is not None:
            self._check_type(minval)
        if maxval is not None:
            self._check_type(maxval)
        self._minval = minval
        self._maxval = maxval
        self._check_bounds(value)

    @property
    def data(self):
        return {
            "value": self.value,
            "value_type": "float",
            "options": self.options,
            "minval": self.minval,
            "maxval": self.maxval,
        }

    @data.setter
    def data(self, data):
        self._value = data["value"]
        self._options = data["options"]
        self._minval = data["minval"]
        self._maxval = data["maxval"]


class IntValue(NumberValue):
    def __init__(self, value=0, minval=None, maxval=None, options=None):
        super().__init__(value, int, options=options)
        if minval is not None:
            self._check_type(minval)
        if maxval is not None:
            self._check_type(maxval)
        self._minval = minval
        self._maxval = maxval
        self._check_bounds(value)

    @property
    def data(self):
        return {
            "value": self.value,
            "value_type": "int",
            "options": self.options,
            "minval": self.minval,
            "maxval": self.maxval,
        }

    @data.setter
    def data(self, data):
        self._value = data["value"]
        self._options = data["options"]
        self._minval = data["minval"]
        self._maxval = data["maxval"]



import scriptcontext as sc  # type: ignore


def register_callback(callback):
    if callback.__name__ not in sc.sticky:
        sc.sticky[callback.__name__] = True
        sc.doc.ReplaceRhinoObject += callback



import os

import Eto.Forms  # type: ignore
import Rhino.UI  # type: ignore
import System  # type: ignore


class FileForm(Eto.Forms.Dialog[bool]):
    @staticmethod
    def open(dirname=None):
        dirname = dirname or os.path.expanduser("~")
        dialog = Eto.Forms.OpenFileDialog()
        if dirname:
            dialog.Directory = System.Uri(dirname)
        dialog.MultiSelect = False
        result = dialog.ShowDialog(Rhino.UI.RhinoEtoApp.MainWindow)
        if result != Eto.Forms.DialogResult.Ok:
            return
        return dialog.FileName

    @staticmethod
    def save(dirname=None, basename=None):
        dirname = dirname or os.path.expanduser("~")
        dialog = Eto.Forms.SaveFileDialog()
        if dirname:
            dialog.Directory = System.Uri(dirname)
        dialog.FileName = basename
        result = dialog.ShowDialog(Rhino.UI.RhinoEtoApp.MainWindow)
        if result != Eto.Forms.DialogResult.Ok:
            return
        return dialog.FileName


class FolderForm(Eto.Forms.Dialog[bool]):
    @staticmethod
    def select(dirname=None):
        dirname = dirname or os.path.expanduser("~")
        dialog = Eto.Forms.SelectFolderDialog()
        dialog.Directory = dirname
        result = dialog.ShowDialog(Rhino.UI.RhinoEtoApp.MainWindow)
        if result != Eto.Forms.DialogResult.Ok:
            return
        return dialog.Directory



from compas.colors import Color

from .value import Value


class ColorValue(Value):
    def __init__(self, value=Color.black()):
        value = Color(*value)
        super(ColorValue, self).__init__(value, Color)

    def cast(self, value):
        try:
            return Color.coerce(value)
        except ValueError:
            raise ValueError("Cannot cast {} to {}".format(value, self.value_type))

    @property
    def data(self):
        return {
            "value": self.value,
            "value_type": "compas.colors.Color",
            "options": self.options,
        }

    @data.setter
    def data(self, data):
        self._value = data["value"]
        self._options = data["options"]



import Eto.Drawing  # type: ignore
import Eto.Forms  # type: ignore
import Rhino  # type: ignore
import Rhino.UI  # type: ignore

from compas.datastructures import Mesh


class MeshInfoForm(Eto.Forms.Dialog[bool]):
    """Form for working with mesh data.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        A COMPAS mesh.
    title : str, optional
        Title of the dialog.
    width : int, optional
        Window width.
    height : int, optional
        Window height.
    vertex_attr_names : list[str]
        The names of vertex attributes to include.
    edge_attr_names : list[str]
        The names of edge attributes to include
    face_attr_names : list[str]
        The names of face attributes to include.

    """

    def __init__(
        self,
        mesh: Mesh,
        vertex_attr_names,
        edge_attr_names,
        face_attr_names,
        title="Mesh Data",
        width=800,
        height=800,
    ):
        super().__init__()

        self.mesh = mesh
        self.vertex_attr_names = vertex_attr_names
        self.edge_attr_names = edge_attr_names
        self.face_attr_names = face_attr_names

        self.Title = title
        self.Padding = Eto.Drawing.Padding(0)
        self.Resizable = True
        self.MinimumSize = Eto.Drawing.Size(0.5 * width, 0.5 * height)
        self.ClientSize = Eto.Drawing.Size(width, height)

        self.vertexpage = Page(
            title="Vertices",
            keys=list(self.mesh.vertices()),
            names=self.vertex_attr_names,
            valuefunc=self.mesh.vertex_attribute,
            defaults=self.mesh.default_vertex_attributes,
        )
        self.edgepage = Page(
            title="Edges",
            keys=list(self.mesh.edges()),
            names=self.edge_attr_names,
            valuefunc=self.mesh.edge_attribute,
            defaults=self.mesh.default_edge_attributes,
        )
        self.facepage = Page(
            title="Faces",
            keys=list(self.mesh.faces()),
            names=self.face_attr_names,
            valuefunc=self.mesh.face_attribute,
            defaults=self.mesh.default_face_attributes,
        )
        control = Eto.Forms.TabControl()
        control.TabPosition = Eto.Forms.DockPosition.Top
        control.Pages.Add(self.vertexpage.widget)
        control.Pages.Add(self.edgepage.widget)
        control.Pages.Add(self.facepage.widget)

        layout = Eto.Forms.DynamicLayout()
        layout.BeginVertical(Eto.Drawing.Padding(0, 12, 0, 12), Eto.Drawing.Size(0, 0), True, True)
        layout.AddRow(control)
        layout.EndVertical()
        layout.BeginVertical(Eto.Drawing.Padding(12, 18, 12, 24), Eto.Drawing.Size(6, 0), False, False)
        layout.AddRow(None, self.ok, self.cancel)
        layout.EndVertical()
        self.Content = layout

    @property
    def ok(self):
        self.DefaultButton = Eto.Forms.Button()
        self.DefaultButton.Text = "OK"
        self.DefaultButton.Click += self.on_ok
        return self.DefaultButton

    @property
    def cancel(self):
        self.AbortButton = Eto.Forms.Button()
        self.AbortButton.Text = "Cancel"
        self.AbortButton.Click += self.on_cancel
        return self.AbortButton

    def on_ok(self, sender, event):
        """Callback for the OK event."""
        self.Close(True)

    def on_cancel(self, sender, event):
        """Callback for the CANCEL event."""
        self.Close(False)

    def show(self):
        """Show the form dialog."""
        return self.ShowModal(Rhino.UI.RhinoEtoApp.MainWindow)


class Page:
    """Wrapper for Eto tab pages.

    Parameters
    ----------
    title : str
        The title of the tab page.
    defaults : dict[str, Any]
        Default values of the data attributes.
    keys : list[int | tuple[int, int]]
        The identifiers of the mesh components.
    valuefunc : callable
        Function for retrieving the value of a named data attribute of a specific component.

    Attributes
    ----------
    names : list[str]
        The names of the attributes, in alphabetical order.
    public : list[str]
        The names of the editable attributes.
    private : list[str]
        The names of the read-only attributes.
    cols : list[dict]
        A list of dicts with each dict representing the properties of a data column.
    rows : list[list[str]]
        The data per mesh components corresponding to the columns.

    """

    def __init__(self, title, keys, names, valuefunc, defaults):
        self.defaults = defaults
        self.keys = keys
        self.names = names
        self.valuefunc = valuefunc
        self.table = Table(self.cols, self.rows)
        self.widget = Eto.Forms.TabPage()
        self.widget.Text = title
        if self.names:
            # replace this by a dynamic layout
            # with the first row an overview of the default attribute values
            # and second row the data table
            self.widget.Content = self.table.widget

    @property
    def cols(self):
        cols = [{"name": "ID", "precision": None}]
        for name in self.names:
            default = self.defaults[name]
            col = {"name": name, "precision": None}
            if isinstance(default, float):
                col["precision"] = "3f"
            cols.append(col)
        return cols

    @property
    def rows(self):
        rows = []
        for key in self.keys:
            row = [str(key)]
            for col in self.cols[1:]:
                name = col["name"]
                precision = col["precision"]
                value = self.valuefunc(key, name)
                if precision:
                    value = "{0:.{1}}".format(float(value), precision)
                else:
                    value = str(value)
                row.append(value)
            rows.append(row)
        return rows


class Table(object):
    """Wrapper for Eto grid view.

    Parameters
    ----------
    cols
    rows

    """

    def __init__(self, cols, rows):
        # def on_cell_formatting(sender, e):
        #     try:
        #         if not e.Column.Editable:
        #             e.ForegroundColor = Eto.Drawing.Colors.DarkGray
        #     except Exception as e:
        #         print(e)

        self.widget = Eto.Forms.GridView()
        self.widget.ShowHeader = True

        for i, col in enumerate(cols):
            column = Eto.Forms.GridColumn()
            column.HeaderText = col["name"]
            column.Editable = False
            cell = Eto.Forms.TextBoxCell(i)
            cell.VerticalAlignment = Eto.Forms.VerticalAlignment.Center
            cell.TextAlignment = Eto.Forms.TextAlignment.Right
            column.DataCell = cell
            self.widget.Columns.Add(column)

        collection = []
        for row in rows:
            item = Eto.Forms.GridItem()
            values = tuple(str(val) if val is not None else "" for val in row)
            item.Values = values
            collection.append(item)

        self.widget.DataStore = collection

        # self.widget.CellFormatting += on_cell_formatting

        self.cols = cols
        self.rows = rows
        self.data = self.widget.DataStore



from .value import Value


class ListValue(Value):
    def __init__(self, value=None, list_value_type=None):
        if value is None:
            value = []
        super(ListValue, self).__init__(value, list)
        self._list_value_type = list_value_type
        self._check_list_value_type(value)

    def _check_list_value_type(self, value):
        for i, item in enumerate(value):
            if self.list_value_type and not isinstance(item, self.list_value_type):
                raise ValueError("List item {}:{} is not of type {}".format(i, item, self.list_value_type))

    def check(self, value):
        super(ListValue, self).check(value)
        self._check_list_value_type(value)

    @property
    def list_value_type(self):
        return self._list_value_type

    @property
    def data(self):
        return {
            "value": self.value,
            "value_type": "list",
            "options": self.options,
            "list_value_type": self.list_value_type.__name__ if self.list_value_type else None,
        }

    @data.setter
    def data(self, data):
        self._value = data["value"]
        self._options = data["options"]



import Rhino.UI  # type: ignore


class Mouse(Rhino.UI.MouseCallback):
    def __init__(self, parent=None):
        super(Mouse, self).__init__()
        self.parent = parent
        self.x = None  # x-coordinate of 2D point in the viewport
        self.y = None  # y-coordinate of 2D point in the viewport
        self.p1 = None  # start of the frustum line in world coordinates
        self.p2 = None  # end of the frustum line in world coordinates

    def OnMouseMove(self, e):
        line = e.View.ActiveViewport.ClientToWorld(e.ViewportPoint)
        self.x = e.ViewportPoint.X
        self.y = e.ViewportPoint.Y
        self.p1 = line.From
        self.p2 = line.To
        e.View.Redraw()

    def OnMouseDown(self, e):
        pass

    def OnMouseUp(self, e):
        pass



import base64
import errno
import json
import os
import uuid
from xml.dom import minidom
from xml.etree import ElementTree as ET

TPL_RUI = """<?xml version="1.0" encoding="utf-8"?>
<RhinoUI major_ver="8"
         minor_ver="0"
         guid="{0}"
         plug_in_guid="{0}"
         localize="False"
         default_language_id="1033">
    <extend_rhino_menus>
        <menu guid="{1}">
          <text>
            <locale_1033>Extend Rhino Menus</locale_1033>
          </text>
        </menu>
    </extend_rhino_menus>
    <menus />
    <tool_bar_groups />
    <tool_bars />
    <macros />
    <bitmaps>
        <small_bitmap item_width="16" item_height="16">
            <bitmap>iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABGdBTUEAALGPC/xhBQAAAAlwSFlz
AAAOvAAADrwBlbxySQAAABNJREFUOE9jGAWjYBSMAjBgYAAABBAAAadEfGMAAAAASUVORK5CYIIA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA==</bitmap>
        </small_bitmap>
        <normal_bitmap item_width="24" item_height="24">
            <bitmap>iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABGdBTUEAALGPC/xhBQAAAAlwSFlz
AAAOvAAADrwBlbxySQAAABNJREFUOE9jGAWjYBSMAjBgYAAABBAAAadEfGMAAAAASUVORK5CYIIA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA==</bitmap>
        </normal_bitmap>
        <large_bitmap item_width="32" item_height="32">
            <bitmap>iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABGdBTUEAALGPC/xhBQAAAAlwSFlz
AAAOvAAADrwBlbxySQAAABNJREFUOE9jGAWjYBSMAjBgYAAABBAAAadEfGMAAAAASUVORK5CYIIA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA==</bitmap>
        </large_bitmap>
    </bitmaps>
    <scripts />
</RhinoUI>
"""

TPL_BITMAP_ITEM = """
<bitmap_item guid="{0}" index="{1}" />
"""

TPL_MACRO_ICON = """
<macro_item guid="{0}" bitmap_id="{7}">
    <text>
        <locale_1033>{1}</locale_1033>
    </text>
    <script>{2}</script>
    <tooltip>
        <locale_1033>{3}</locale_1033>
    </tooltip>
    <help_text>
        <locale_1033>{4}</locale_1033>
    </help_text>
    <button_text>
        <locale_1033>{5}</locale_1033>
    </button_text>
    <menu_text>
        <locale_1033>{6}</locale_1033>
    </menu_text>
</macro_item>
"""

TPL_MACRO = """
<macro_item guid="{0}">
    <text>
        <locale_1033>{1}</locale_1033>
    </text>
    <script>{2}</script>
    <tooltip>
        <locale_1033>{3}</locale_1033>
    </tooltip>
    <help_text>
        <locale_1033>{4}</locale_1033>
    </help_text>
    <button_text>
        <locale_1033>{5}</locale_1033>
    </button_text>
    <menu_text>
        <locale_1033>{6}</locale_1033>
    </menu_text>
</macro_item>
"""

TPL_MENUITEM = """
<menu_item guid="{0}" item_type="normal">
    <macro_id>{1}</macro_id>
</menu_item>
"""

TPL_MENUSEPARATOR = """
<menu_item guid="{0}" item_type="separator"></menu_item>
"""

TPL_TOOLBARITEM = """
<tool_bar_item guid="{0}" button_style="normal">
    <left_macro_id>{1}</left_macro_id>
    <right_macro_id>{2}</right_macro_id>
</tool_bar_item>
"""

TPL_TOOLBARSEPARATOR = """
<tool_bar_item guid="{0}" button_display_mode="control_only" button_style="spacer">
</tool_bar_item>
"""

TPL_TOOLBAR = """
<tool_bar guid="{0}" item_display_style="{2[item_display_style]}">
    <text>
        <locale_1033>{1}</locale_1033>
    </text>
</tool_bar>
"""

TPL_TOOLBARGROUP = """
<tool_bar_group guid="{0}"
                dock_bar_guid32="{0}"
                dock_bar_guid64="{0}"
                active_tool_bar_group=""
                single_file="{2[single_file]}"
                hide_single_tab="{2[hide_single_tab]}"
                point_floating="0,0">
    <text>
        <locale_1033>{1}</locale_1033>
    </text>
</tool_bar_group>
"""

TPL_TOOLBARGROUPITEM = """
<tool_bar_group_item guid="{0}" major_version="1" minor_version="1">
    <text>
        <locale_1033>{1}</locale_1033>
    </text>
    <tool_bar_id>{2}</tool_bar_id>
</tool_bar_group_item>
"""


class Rui(object):
    """Class for generating *.rui files.

    Parameters
    ----------
    filepath : str
        Path to the *.rui file.

    """

    def __init__(self, filepath, guid=None):
        self.filepath = filepath
        self.guid = guid or uuid.uuid4()
        self.icons = []
        self.macros = {}
        self.toolbars = {}
        self.xml = None
        self.root = None
        self.root_bitmaps = []
        self.root_macros = []
        self.root_menus = []
        self.root_toolbargroups = []
        self.root_toolbars = []
        self.check()

    def check(self):
        if not os.path.exists(os.path.dirname(self.filepath)):
            try:
                os.makedirs(os.path.dirname(self.filepath))
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise e
        if not os.path.exists(self.filepath):
            with open(self.filepath, "w+"):
                pass

    def init(self):
        with open(self.filepath, "w+") as f:
            f.write(TPL_RUI.format(self.guid, self.guid))
        self.xml = ET.parse(self.filepath)
        self.root = self.xml.getroot()
        self.root_bitmaps = self.root.find("bitmaps")
        self.root_macros = self.root.find("macros")
        self.root_menus = self.root.find("menus")
        self.root_toolbargroups = self.root.find("tool_bar_groups")
        self.root_toolbars = self.root.find("tool_bars")

    def parse(self):
        raise NotImplementedError

    def write(self):
        root = ET.tostring(self.root)  # type: ignore
        xml = minidom.parseString(root).toprettyxml(indent="  ")
        xml = "\n".join([line for line in xml.split("\n") if line.strip()])
        with open(self.filepath, "w+") as fh:
            fh.write(xml)

    @classmethod
    def from_json(cls, uipath, ruipath, guid=None) -> "Rui":
        """Construct a RUI object from a JSON config file.

        Parameters
        ----------
        uipath : str
            Path to the UI config file.
        ruipath : str
            Path of the output file.
        guid : uuid, optional
            The guid of the plugin.

        Returns
        -------
        :class:`Rui`

        """
        root = os.path.dirname(uipath)

        with open(uipath, "r") as f:
            ui = json.load(f)

        self = cls(ruipath, guid=guid)
        self.init()
        if ui["icons"]["bitmap"]:
            bitmap = os.path.join(root, ui["icons"]["bitmap"])
            if os.path.exists(bitmap):
                self.add_bitmap(bitmap)
                self.add_bitmap_items(ui["icons"]["images"])
        self.add_macros(ui["commands"])
        self.add_menus(ui["menus"])
        self.add_toolbars(ui["toolbars"])
        self.add_toolbargroups(ui["toolbargroups"])

        return self

    # --------------------------------------------------------------------------
    # add icons
    # --------------------------------------------------------------------------

    def add_bitmap(self, path):
        with open(path, "rb") as f:
            bitmap = base64.b64encode(f.read()).decode("utf-8")
        self.root_bitmaps.find("large_bitmap").find("bitmap").text = bitmap

    def add_bitmap_items(self, items):
        small = self.root_bitmaps.find("small_bitmap")
        normal = self.root_bitmaps.find("normal_bitmap")
        large = self.root_bitmaps.find("large_bitmap")
        for index, name in enumerate(items):
            guid = str(uuid.uuid4())
            s = TPL_BITMAP_ITEM.format(guid, index)
            e = ET.fromstring(s)
            small.append(e)
            normal.append(e)
            large.append(e)
            self.icons.append(guid)

    # --------------------------------------------------------------------------
    # add macros
    # --------------------------------------------------------------------------

    def add_macros(self, macros):
        for macro in macros:
            guid = str(uuid.uuid4())
            name = macro["name"]
            script = macro.get("script") or "-_{}".format(name)
            tooltip = macro.get("tooltip", "")
            help_text = macro.get("help_text", "")
            button_text = macro.get("button_text", name)
            menu_text = macro.get("menu_text", name.replace("_", " "))
            icon_index = macro.get("icon")
            self.add_macro(
                name,
                guid,
                script,
                tooltip,
                help_text,
                button_text,
                menu_text,
                icon_index,
            )

    def add_macro(
        self,
        name,
        guid,
        script,
        tooltip,
        help_text,
        button_text,
        menu_text,
        icon_index=None,
    ):
        if icon_index is not None:
            icon_guid = self.icons[icon_index]
            s_macro = TPL_MACRO_ICON.format(
                guid,
                name,
                script,
                tooltip,
                help_text,
                button_text,
                menu_text,
                icon_guid,
            )
        else:
            s_macro = TPL_MACRO.format(guid, name, script, tooltip, help_text, button_text, menu_text)
        e_macro = ET.fromstring(s_macro)
        self.root_macros.append(e_macro)
        self.macros[name] = e_macro

    # --------------------------------------------------------------------------
    # add menus
    # --------------------------------------------------------------------------

    def add_menus(self, menus):
        for menu in menus:
            self.add_menu(menu)

    def add_menu(self, menu, root=None):
        if root is None:
            root = self.root_menus
        e_menu = ET.SubElement(root, "menu")
        e_menu.set("guid", str(uuid.uuid4()))
        e_text = ET.SubElement(e_menu, "text")
        e_locale = ET.SubElement(e_text, "locale_1033")
        e_locale.text = menu["name"]
        for item in menu["items"]:
            if "command" in item:
                e_macro = self.macros[item["command"]]
                macro_guid = e_macro.attrib["guid"]
                self.add_menuitem(e_menu, macro_guid)
            elif "items" in item:
                self.add_menu(item, root=e_menu)
            elif item.get("type") == "separator":
                self.add_menuseparator(e_menu)

    def add_menuitem(self, root, macro_id):
        guid = uuid.uuid4()
        s_item = TPL_MENUITEM.format(guid, macro_id)
        e_item = ET.fromstring(s_item)
        root.append(e_item)

    def add_menuseparator(self, root):
        guid = uuid.uuid4()
        s_sep = TPL_MENUSEPARATOR.format(guid)
        e_sep = ET.fromstring(s_sep)
        root.append(e_sep)

    # --------------------------------------------------------------------------
    # add toolbars
    # --------------------------------------------------------------------------

    def add_toolbars(self, toolbars):
        for toolbar in toolbars:
            self.add_toolbar(toolbar)

    def add_toolbar(self, toolbar):
        options = {"item_display_style": "control"}
        guid = uuid.uuid4()
        s_tb = TPL_TOOLBAR.format(guid, toolbar["name"], options)
        e_tb = ET.fromstring(s_tb)
        self.root_toolbars.append(e_tb)
        self.toolbars[toolbar["name"]] = e_tb
        for item in toolbar["items"]:
            if item.get("type") == "separator":
                self.add_toolbarseparator(e_tb)
                continue
            left_guid = None
            left_macro = item.get("left")
            if left_macro:
                e_left = self.macros[left_macro]
                left_guid = e_left.attrib["guid"]
            right_guid = None
            right_macro = item.get("right")
            if right_macro:
                e_right = self.macros[right_macro]
                right_guid = e_right.attrib["guid"]
            self.add_toolbaritem(e_tb, left_guid, right_guid)

    def add_toolbaritem(self, root, left_macro_id, right_macro_id):
        guid = uuid.uuid4()
        s_item = TPL_TOOLBARITEM.format(guid, left_macro_id, right_macro_id)
        e_item = ET.fromstring(s_item)
        root.append(e_item)

    def add_toolbarseparator(self, root):
        guid = uuid.uuid4()
        s_sep = TPL_TOOLBARSEPARATOR.format(guid)
        e_sep = ET.fromstring(s_sep)
        root.append(e_sep)

    # --------------------------------------------------------------------------
    # add toolbargroups
    # --------------------------------------------------------------------------

    def add_toolbargroups(self, toolbargroups):
        for tbg in toolbargroups:
            self.add_toolbargroup(tbg)

    def add_toolbargroup(self, tbg):
        options = {
            "single_file": "False",
            "hide_single_tab": "False",
        }
        guid = uuid.uuid4()
        s_tbg = TPL_TOOLBARGROUP.format(guid, tbg["name"], options)
        e_tbg = ET.fromstring(s_tbg)
        self.root_toolbargroups.append(e_tbg)
        for tb_name in tbg["toolbars"]:
            e_tb = self.toolbars[tb_name]
            tb_guid = e_tb.attrib["guid"]
            self.add_toolbargroupitem(e_tbg, tb_name, tb_guid)

    def add_toolbargroupitem(self, root, tb_name, tb_guid):
        guid = uuid.uuid4()
        s_item = TPL_TOOLBARGROUPITEM.format(guid, tb_name, tb_guid)
        e_item = ET.fromstring(s_item)
        root.append(e_item)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="COMPAS Rhino UI compiler.")

    parser.add_argument("ipath", help="Path to the config file (input).")
    parser.add_argument("opath", help="Path to the rui file (output).")

    args = parser.parse_args()

    ipath = os.path.abspath(args.ipath)
    opath = os.path.abspath(args.opath)

    rui = Rui.from_json(ipath, opath)
    rui.write()
